import json
import logging
from contextlib import suppress
from datetime import datetime
from typing import Dict, List

import redis

from .api import _APIClient
from .config import Client
from .constants import _TMIO
from .errors import InvalidIDError, TMIOException

_log = logging.getLogger(__name__)

__all__ = (
    "PlayerMatchmakingResult",
    "PlayerMatchmaking",
)


async def _get_top_matchmaking(page: int = 0, royal: bool = False):
    _log.debug(f"Getting top matchmaking players page {page}. Royal? {royal}")

    cache_client = Client._get_cache_client()

    with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
        if cache_client.exists(f"top_matchmaking:{page}:{royal}"):
            _log.debug(f"Found top matchmaking players for page {page} in cache")
            return json.loads(
                cache_client.get(f"top_matchmaking:{page}:{royal}").decode("utf-8")
            ).get("ranks")
    api_client = _APIClient()

    if not royal:
        match_history = await api_client.get(
            _TMIO.build([_TMIO.TABS.TOP_MATCHMAKING, str(page)])
        )
    else:
        match_history = await api_client.get(
            _TMIO.build([_TMIO.TABS.TOP_ROYAL, str(page)])
        )

    await api_client.close()

    with suppress(KeyError, TypeError):
        raise TMIOException(match_history["error"])
    with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
        _log.debug(f"Caching top matchmaking players for page {page}")
        cache_client.set(
            f"top_matchmaking:{page}:{royal}", json.dumps(match_history), ex=3600
        )

    return match_history.get("ranks")


async def _get_history(player_id: str, type_id: int, page: int) -> List[Dict]:
    if player_id is None:
        raise InvalidIDError("Player ID is not set.")

    _log.debug("Getting matchmaking history for player %s and page %d", player_id, page)

    cache_client = Client._get_cache_client()

    with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
        if cache_client.exists(f"mm_hist:{page}:{type_id}:{player_id}"):
            _log.debug("Found matchmaking history for page %s in cache", page)
            return json.loads(
                cache_client.get(f"mm_hist:{page}:{type_id}:{player_id}").decode(
                    "utf-8"
                )
            )["history"]

    api_client = _APIClient()
    match_history = await api_client.get(
        _TMIO.build(
            [
                _TMIO.TABS.PLAYER,
                player_id,
                _TMIO.TABS.MATCHES,
                type_id,
                page,
            ]
        )
    )
    await api_client.close()

    with suppress(KeyError, TypeError):
        raise TMIOException(match_history["error"])
    with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
        _log.debug(f"Saving matchmaking history for page {page} to cache")
        cache_client.set(
            f"mm_history:{page}:{type_id}:{player_id}",
            json.dumps(match_history),
            ex=3600,
        )

    return match_history.get("history", [])


class PlayerMatchmakingResult:
    """
    .. versionadded :: 0.3.0

    Represent's a player's matchmaking result

    Parameters
    ----------
    after_score : int
        The player's matchmaking score after the match
    leave : bool
        Whether the player left the match
    live_id : str
        The live id of the match
    mvp : bool
        Whether the player was the mvp of the match
    player_id : str
        The player's ID
    start_time : :class:`datetime`
        The date the match started
    win : bool
        Whether the player won the match
    """

    def __init__(
        self,
        after_score: int,
        leave: bool,
        live_id: str,
        mvp: bool,
        player_id: str | None,
        start_time: datetime,
        win: bool,
    ):
        self.after_score = after_score
        self.leave = leave
        self.live_id = live_id
        self.mvp = mvp
        self.player_id = player_id
        self.start_time = start_time
        self.win = win

    @classmethod
    def _from_dict(cls, data: Dict, player_id: str = None):
        _log.debug("Creating a PlayerMatchmakingResult class from given dictionary")

        after_score = data.get("afterscore")
        leave = data.get("leave")
        live_id = data.get("lid")
        mvp = data.get("mvp")
        start_time = datetime.strptime(data.get("startime"), "%Y-%m-%dT%H:%M:%SZ")
        win = data.get("win")

        args = [after_score, leave, live_id, mvp, player_id, start_time, win]

        return cls(*args)


class PlayerMatchmaking:
    """
    .. versionadded :: 0.1.0

    Class that represents the player matchmaking details

    Parameters
    ----------
    type : str
        The type of matchmaking, either "3v3" or "Royal"
    type_id : int
        The type of matchmaking as 0 or 1, 0 for "3v3" and 1 for "Royal"
    progression : int
        The progression of the player's score in matchmaking
    rank : int
        The rank of the player in matchmaking
    score : int
        The score of the player in matchmaking
    division : int
        The division of the player in matchmaking
    division_str : str: str
        The division of the player in matchmaking as a string
    min_points : int
        The points required to reach the current division.
    max_points : int
        The points required to move up the rank.
    player_id : str | None
        The player's ID. Defaults to None
    """

    def __init__(
        self,
        matchmaking_type: str,
        type_id: int,
        progression: int,
        rank: int,
        score: int,
        division: int,
        min_points: int,
        max_points: int,
        player_id: str | None = None,
    ):
        """Constructor for the class."""
        MATCHMAKING_STRING = {
            1: "Bronze 3",
            2: "Bronze 2",
            3: "Bronze 1",
            4: "Silver 3",
            5: "Silver 2",
            6: "Silver 1",
            7: "Gold 3",
            8: "Gold 2",
            9: "Gold 1",
            10: "Master 3",
            11: "Master 2",
            12: "Master 1",
            13: "Trackmaster",
        }

        self.matchmaking_type = matchmaking_type
        self.type_id = type_id
        self.rank = rank
        self.score = score
        self.progression = progression
        self.division = division
        self.division_str = MATCHMAKING_STRING.get(division)
        self._min_points = min_points
        self._max_points = 1 if max_points == 0 else max_points
        self.player_id = player_id

        try:
            self.progress = round(
                (score - min_points) / (max_points - min_points) * 100, 2
            )
        except ZeroDivisionError:
            self.progress = 0

    @staticmethod
    def _from_dict(mm_data: Dict, player_id: str = None):
        """
        Parses the matchmaking data of the player and returns 2 :class:`PlayerMatchmaking` objects.
            One for 3v3 Matchmaking and the other for Royal matchmaking.

        Parameters
        ----------
        mm_data : :class:`List[Dict]`
            The matchmaking data.
        player_id : str, optional
            The player's ID. Defaults to None
        Returns
        -------
        :class:`List[PlayerMatchmaking]`
            The list of matchmaking data, one for 3v3 and other other one for royal.
        """
        _log.debug("Creating a PlayerMatchmaking class from given dictionary")

        matchmaking_data = []

        if len(mm_data) == 0:
            matchmaking_data.extend([None, None])
        elif len(mm_data) == 1:
            mm_obj = PlayerMatchmaking.__parse_3v3(mm_data[0])
            matchmaking_data.extend(
                [mm_obj, None] if mm_obj.type_id == 2 else [None, mm_obj]
            )
        else:
            matchmaking_data.extend(
                [
                    PlayerMatchmaking.__parse_3v3(mm_data[0], player_id),
                    PlayerMatchmaking.__parse_3v3(mm_data[1], player_id),
                ]
            )

        return matchmaking_data

    @classmethod
    def __parse_3v3(cls, data: Dict, player_id: str = None):
        """
        Parses matchmaking data for 3v3 and royal type matchmaking.

        Parameters
        ----------
        data : :class:`Dict`
            The matchmaking data only.
        Returns
        -------
        :class:`PlayerMatchmaking`
            The parsed data.
        """
        _log.debug(
            f"Parsing Data from Dictionary for PlayerMatchmaking class. ID supplied: {player_id}"
        )

        if "info" in data:
            data = data.get("info")

        type_name = data.get("typename")
        type_id = data.get("typeid")
        progression = data.get("progression")
        rank = data.get("rank")
        score = data.get("score")
        division = data.get("division").get("position")
        min_points = data.get("division").get("minpoints")
        max_points = data.get("division").get("maxpoints")

        args = [
            type_name,
            type_id,
            progression,
            rank,
            score,
            division,
            min_points,
            max_points,
            player_id,
        ]

        return cls(*args)

    @property
    def min_points(self):
        """min points"""
        return self._min_points

    @property
    def max_points(self):
        """max points"""
        return self._max_points

    def __str__(self):
        progression = self.progression
        progress = self.progress
        rank = self.rank
        score = self.score
        division = self.division
        division_str = self.division_str
        max_points = self.max_points

        return f"Progression: {progression}\nProgress: {progress}\nRank: {rank}\nScore: {score}\nDivision: {division_str} - {division}\n\nPoints to Next Division: {max_points + 1}"

    async def history(self, page: int = 0) -> List[PlayerMatchmakingResult]:
        """
        .. versionadded :: 0.3.0
        .. versionchanged :: 0.4.0
            Use `_get_history()` helper command.

        History of recent matches in this matchmaking

        Parameters
        ----------
        page : int, optional
            The page number, by default 0

        Returns
        -------
        :class:`List[PlayerMatchmakingResult]`
            The list of matchmaking results

        Raises
        ------
        :class:`InvalidIDError`
            If the player_id is not set.
        """
        matches = await _get_history(self.player_id, self.type_id, page)

        match_results = []
        for match in matches:
            match_results.append(
                PlayerMatchmakingResult._from_dict(match), self.player_id
            )

        return match_results

    @staticmethod
    async def top_matchmaking(page: int = 0, royal: bool = False) -> List[Dict]:
        """
        .. versionadded :: 0.3.0

        Top matchmaking players

        Parameters
        ----------
        page : int, optional
            The page number, by default 0
        royal : bool, optional
            Whether to get the top matchmaking players for royal, by default False

        Returns
        -------
        :class:`List[Dict]`
            The top matchmaking players by score. Each page contains 50 players.
        """
        return await _get_top_matchmaking(page, royal)



import json
import logging
from contextlib import suppress
from datetime import datetime
from typing import Dict, List

import redis
from typing_extensions import Self

from .api import _APIClient
from .config import Client
from .constants import _TMIO
from .errors import TMIOException
from .matchmaking import PlayerMatchmaking
from .trophy import PlayerTrophies

_log = logging.getLogger(__name__)

__all__ = (
    "PlayerMetaInfo",
    "PlayerZone",
    "PlayerMatchmaking",
    "Player",
)


class PlayerMetaInfo:
    """
    .. versionadded :: 0.1.0

    Represents Player Meta Data, which inclues YT, Twitch, Twitter or _TMIO Vanity Link

    Parameters
    ----------
    display_url : str
        The URL to the player's profile
    in_nadeo : bool
        Whether the player is in Nadeo
    in_tmgl : bool
        Whether the player is in TMGL
    in_tmio_dev_team : bool
        Whether the player is in _TMIO Dev Team
    is_sponsor : bool
        Whether the player is a sponsor
    sponsor_level : int | None
        The sponsor level of the player
    twitch : str | None
        The Twitch URL of the player, `NoneType` if the player has no Twitch
    twitter : str | None
        The Twitter URL of the player, `NoneType` if the player has no Twitter
    youtube : str | None
        The YouTube URL of the player, `NoneType` if the player has no YouTube
    vanity : str | None
        The _TMIO Vanity URL of the player, `NoneType` if the player has no _TMIO Vanity URL
    """

    def __init__(
        self,
        display_url: str,
        in_nadeo: bool,
        in_tmgl: bool,
        in_tmio_dev_team: bool,
        is_sponsor: bool,
        sponsor_level: int | None,
        twitch: str | None,
        twitter: str | None,
        youtube: str | None,
        vanity: str | None,
    ):
        """
        Constructor method.
        """
        self.display_url = display_url
        self.in_nadeo = in_nadeo
        self.in_tmgl = in_tmgl
        self.in_tmio_dev_team = in_tmio_dev_team
        self.is_sponsor = is_sponsor
        self.sponsor_level = sponsor_level
        self.twitch = twitch
        self.twitter = twitter
        self.youtube = youtube
        self.vanity = vanity

    @classmethod
    def _from_dict(cls, meta_data: Dict):
        """
        .. versionadded :: 0.1.0

        Parses the meta data into a PlayerMetaInfo object.

        Parameters
        ----------
        meta_data : Dict
            The meta data to parse
        Returns
        -------
        :class:`PlayerMetaInfo`
            The parsed meta data
        """
        _log.debug(f"Creating a PlayerMetaInfo class from the given dictionary.")

        return cls(
            display_url=meta_data.get("displayurl"),
            in_nadeo=meta_data.get("nadeo", False),
            in_tmgl=meta_data.get("tmgl", False),
            in_tmio_dev_team=meta_data.get("team", False),
            is_sponsor=meta_data.get("sponsor", False),
            sponsor_level=meta_data.get("sponsor_level", 0),
            twitch=meta_data.get("twitch"),
            twitter=meta_data.get("twitter"),
            youtube=meta_data.get("youtube"),
            vanity=meta_data.get("vanity"),
        )


class PlayerZone:
    """
    .. versionadded :: 0.1.0

    Class that represents the player zone

    Parameters
    ----------
    flag : str
        The flag of the zone
    zone : str
        The zone name
    rank : int
        The rank of the player in the zone
    """

    def __init__(self, flag: str, zone: str, rank: int):
        """Constructor method."""
        self.flag = flag
        self.zone = zone
        self.rank = rank

    @classmethod
    def _parse_zones(cls: Self, zones: Dict, zone_positions: List[int]) -> List[Self]:
        """
          .. versionadded :: 0.1.0

        Parses the Data from the API into a list of PlayerZone objects.

         Parameters
         ----------
         zones : :class:`Dict`
             the zones data from the API.
         zone_positions : :class:`List[int]`
             The zone positions data from the API.
         Returns
         -------
         class:`List[PlayerZone]`
             The list of :class:`PlayerZone` objects.
        """
        _log.debug("Parsing Zones")
        player_zone_list: List = []
        i: int = 0

        while "name" in zones:
            _log.debug(f"Gone {i} Levels Deep")
            player_zone_list.append(
                cls(zones["flag"], zones["name"], zone_positions[i])
            )
            i += 1

            if "parent" in zones:
                zones = zones["parent"]
            else:
                break

        return player_zone_list

    @staticmethod
    def to_string(
        player_zones: List[Self] | None, add_pos: bool = True, inline: bool = False
    ) -> str | None:
        """
        .. versionadded :: 0.4.0

        Prints a list of zones in a readable format.

        Parameters
        ----------
        player_zones : :class:`List[Self]`
            The list of :class:`PlayerZone` objects.
        add_pos : bool
            .. versionadded:: 0.4.0
            Whether to add the position of the zone.
        inline : bool
            .. versionadded :: 0.4.0
            Whether to print the zones in a single line.

        Returns
        -------
        str
            The list of zones in a readable format.
        """
        if player_zones is None:
            return None

        zone_str = ""

        if not inline:
            if add_pos:
                for zone in player_zones:
                    zone_str = zone_str + zone.zone + " - " + str(zone.rank) + "\n"
            else:
                for zone in player_zones:
                    zone_str = zone_str + zone.zone + "\n"
        else:
            if add_pos:
                zone_str = ", ".join(
                    f"{zone.zone} - {zone.rank}" for zone in player_zones
                )
            else:
                zone_str = ", ".join(zone.zone for zone in player_zones)

        return zone_str


class PlayerSearchResult:
    """
    .. versionadded :: 0.1.0

    Represents 1 Player from a Search Result

    Parameters
    ----------
    club_tag : str | None.
        The club tag of the player, `NoneType` if the player is not in a club.
    name : str
        Name of the player.
    player_id : str
        The Trackmania ID of the player.
    zone : :class:`List[PlayerZone]`, optional
        The zone of the player as a list.
    threes : :class:`PlayerMatchmaking`, optional
        The 3v3 data of the player.
    royals : :class:`PlayerMatchmaking`, optional
        The royal data of the player.
    """

    def __init__(
        self,
        club_tag: str | None,
        name: str,
        player_id: str,
        zone: List[PlayerZone],
        threes: PlayerMatchmaking | None,
        royal: PlayerMatchmaking | None,
    ):
        self.club_tag = club_tag
        self.name = name
        self.player_id = player_id
        self.zone = zone
        self.threes = threes
        self.royal = royal

    @classmethod
    def _from_dict(cls: Self, player_data: Dict) -> Self:
        _log.debug("Creating a PlayerSearchResult class from given dictionary")

        zone = (
            PlayerZone._parse_zones(player_data["player"]["zone"], [0, 0, 0, 0, 0])
            if "zone" in player_data["player"]
            else None
        )
        club_tag = player_data.get("player").get("club_tag", None)
        name = player_data.get("player").get("name")
        player_id = player_data.get("player").get("id")
        matchmaking = PlayerMatchmaking._from_dict(
            player_data.get("matchmaking"), player_id
        )

        return cls(club_tag, name, player_id, zone, matchmaking[0], matchmaking[1])


class Player:
    """
    .. versionadded :: 0.1.0

    Represents a Player in Trackmania

    Parameters
    ----------
    club_tag : str | None.
        The club tag of the player, `NoneType` if the player is not in a club.
    first_login : :class:`datetime` | None
        The date of the first login of the player.
    player_id : str
        The Trackmania ID of the player.
    last_club_tag_change : str
        The date of the last club tag change of the player.
    login : str
        Login of the player.
    meta : :class:`PlayerMetaInfo`.
        Meta data of the player.
    name : str
        Name of the player.
    trophies : :class:`PlayerTrophies`, optional
        The trophies of the player.
    zone : :class:`List[PlayerZone]`, optional
        The zone of the player as a list.
    m3v3_data : :class:`PlayerMatchmaking`, optional
        The 3v3 data of the player.
    royal_data : :class:`PlayerMatchmaking`, optional
        The royal data of the player.
    """

    def __init__(
        self,
        club_tag: str | None,
        first_login: datetime | None,
        player_id: str,
        last_club_tag_change: str,
        meta: PlayerMetaInfo,
        name: str,
        trophies: PlayerTrophies | None = None,
        zone: List[PlayerZone] | None = None,
        m3v3_data: PlayerMatchmaking | None = None,
        royal_data: PlayerMatchmaking | None = None,
    ):
        """Constructor of the class."""
        self.club_tag = club_tag
        self._first_login = first_login
        self._id = player_id
        self.last_club_tag_change = last_club_tag_change
        self.meta = meta
        self.name = name
        self.trophies = trophies
        self.zone = zone
        self.m3v3_data = m3v3_data
        self.royal_data = royal_data

    def __str__(self):
        """String representation of the class."""
        return f"Player: {self.name} ({self.player_id})"

    @property
    def first_login(self):
        """first login property."""
        return self._first_login

    @property
    def player_id(self):
        """player id property."""
        return self._id

    @classmethod
    async def get_player(cls: Self, player_id: str) -> Self:
        """
        .. versionadded :: 0.1.0

        Gets a player's data from their player_id

        Parameters
        ----------
        player_id : str
            The player id of the player
        """
        _log.debug(f"Getting {player_id}'s data")

        cache_client = Client._get_cache_client()

        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            if cache_client.exists(f"player:{player_id}"):
                _log.debug(f"{player_id}'s data found in cache")
                player_data = cache_client.get(f"player:{player_id}").decode("utf-8")
                player_data = json.loads(player_data)
                return cls(
                    **Player._parse_player(
                        json.loads(
                            cache_client.get(f"player:{player_id}").decode("utf-8")
                        )
                    )
                )

        api_client = _APIClient()
        player_data = await api_client.get(_TMIO.build([_TMIO.TABS.PLAYER, player_id]))
        await api_client.close()

        with suppress(KeyError, TypeError):

            raise TMIOException(player_data["error"])
        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            cache_client.set(f"player:{player_id}", json.dumps(player_data))
            cache_client.set(f"{player_data['displayname'].lower()}:id", player_id)

        return cls(**Player._parse_player(player_data))

    @staticmethod
    async def search(
        username: str,
    ) -> List[PlayerSearchResult] | None:
        """
        .. versionadded :: 0.1.0
        .. versionchanged :: 0.3.4
            The function no longer returns a single :class:`PlayerSearchResult`. It will now always return a `list` or `None`

        Searches for a player's information using their username.

        Parameters
        ----------
        username : str
            The player's username to search for

        Returns
        -------
        :class:`List[PlayerSearchResult]` | None
            Returns a list of :class:`PlayerSearchResult` with users who have similar usernames. Returns `None`
            if no user with that username can be found.
        """
        _log.debug(f"Searching for players with the username -> {username}")

        api_client = _APIClient()
        search_result = await api_client.get(
            _TMIO.build([_TMIO.TABS.PLAYERS]) + f"/find?search={username}"
        )
        await api_client.close()

        with suppress(KeyError, TypeError):
            raise TMIOException(search_result["error"])

        if len(search_result) == 0:
            return None

        players = []
        for player in search_result:
            players.append(PlayerSearchResult._from_dict(player))

        return players

    @staticmethod
    async def get_id(username: str) -> str:
        """
        .. versionadded :: 0.1.0
        .. versionadded :: 0.3.4
            Updated to work with the change in `search` function

        Gets a player's id from the given username

        Parameters
        ----------
        username : str
            The player's username to get the ID for.

        Returns
        -------
        str
            The player's id.
        """
        _log.debug(f"Getting {username}'s id")

        cache_client = Client._get_cache_client()

        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            if cache_client.exists(f"{username.lower()}:id"):
                _log.debug(f"{username}'s id found in cache")
                return cache_client.get(f"{username.lower()}:id").decode("utf-8")

        players = await Player.search(username)

        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            _log.debug(f"Caching {username.lower()} id as {players[0].player_id}")
            cache_client.set(f"{username.lower()}:id", players[0].player_id)
        return players[0].player_id

    @staticmethod
    async def get_username(player_id: str) -> str:
        """
        .. versionadded :: 0.1.0

        Gets a player's username from their player id

        Parameters
        ----------
        player_id : str
            The player id of the player

        Returns
        -------
        str
            The player's username
        """
        _log.debug(f"Getting the username for {player_id}")
        cache_client = Client._get_cache_client()

        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            if cache_client.exists(f"{player_id}:username"):
                _log.debug(f"{player_id}'s username found in cache")
                return json.loads(
                    cache_client.get(f"{player_id}:username").decode("utf-8")
                )

        player: Player = await Player.get_player(player_id)

        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            _log.debug(f"Caching {player_id}:username as {player.name}")
            cache_client.set(f"{player_id}:username", player.name)

        return player.name

    @staticmethod
    def _parse_player(player_data: Dict) -> Dict:
        """
        .. versionadded :: 0.1.0
        .. versionchanged :: 0.4.0
            Optimized everything!

        Parses the player data

        Parameters
        ----------
        player_data : :class:`Dict`
            The player data as a dictionary

        Returns
        -------
        :class:`Dict`
            The parsed player data formatted kwargs friendly for the :class:`Player` constructors
        """
        # Parsing First Login
        first_login = (
            datetime.strptime(player_data["timestamp"], "%Y-%m-%dT%H:%M:%S+00:00")
            if "timestamp" in player_data
            else None
        )

        # Parsing Last Club Tag Change
        last_club_tag_change = (
            datetime.strptime(
                player_data["clubtagtimestamp"], "%Y-%m-%dT%H:%M:%S+00:00"
            )
            if "clubtagtimestamp" in player_data
            else None
        )

        # Parsing Meta
        if player_data.get("meta") is not None:
            if isinstance(player_data["meta"], PlayerMetaInfo):
                player_meta = player_data.get("meta")
            else:
                player_meta = PlayerMetaInfo._from_dict(player_data.get("meta"))
        else:
            player_meta = PlayerMetaInfo._from_dict(dict())

        # Parsing Trophies
        player_trophies = player_data.get("trophies")
        if player_trophies is not None:
            player_trophies = PlayerTrophies._from_dict(
                player_trophies,
                player_data.get("accountid", player_data.get("player_id")),
            )

        # Parsing Zones
        if (
            player_trophies is not None
            and player_data.get("trophies").get("zone") is not None
        ):
            player_zone = PlayerZone._parse_zones(
                player_data.get("trophies").get("zone"),
                player_data.get("trophies").get("zonepositions"),
            )
        else:
            player_zone = False

        # Parsing player id
        player_id = player_data.get(
            "accountid", player_data.get("id", player_data.get("playerid", None))
        )

        # Parsing Matchmaking
        matchmaking = (
            PlayerMatchmaking._from_dict(player_data["matchmaking"], player_id)
            if "matchmaking" in player_data
            else [None, None]
        )

        # Parsing Club Tag
        club_tag = player_data.get("clubtag", player_data.get("tag", None))

        # Parsing Name
        name = player_data.get("displayname", player_data.get("name", None))

        return {
            "club_tag": club_tag,
            "first_login": first_login,
            "name": name,
            "player_id": player_id,
            "last_club_tag_change": last_club_tag_change,
            "meta": player_meta,
            "trophies": player_trophies,
            "zone": player_zone,
            "m3v3_data": matchmaking[0],
            "royal_data": matchmaking[1],
        }



import json
import logging
from contextlib import suppress
from typing import Dict, List

import redis
from typing_extensions import Self

from trackmania.errors import TMIOException

from .api import _APIClient
from .config import Client
from .constants import _TMIO

_log = logging.getLogger(__name__)

# __all__ = ("Ad", "list_ads", "get_ad")
__all__ = ("Ad",)


async def _get_ad_list() -> List[Dict]:
    ad_list = []

    _log.debug("Getting all ads")
    cache_client = Client._get_cache_client()
    with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
        if cache_client.exists("ads"):
            _log.debug("Found all ads in cache")
            ads = json.loads(cache_client.get("ads").decode("utf-8"))
            for ad_dict in ads.get("ads"):
                ad_list.append(ad_dict)
            return ad_list

    api_client = _APIClient()
    all_ads = await api_client.get(_TMIO.build([_TMIO.TABS.ADS]))
    await api_client.close()

    with suppress(KeyError, TypeError):
        raise TMIOException(all_ads["error"])
    with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
        _log.debug("Caching all ads for 12hours")
        cache_client.set("ads", json.dumps(all_ads), ex=43200)

    for ad_dict in all_ads.get("ads"):
        ad_list.append(ad_dict)

    return ad_list


class Ad:
    """
    .. versionadded :: 0.3.0

    Represents an Ad in Trackmania

    Parameters
    ----------
    uid : str
        The unique ID of the ad
    name : str
        The name of the ad
    type : str
        The type of the ad
    url : str
        URL set for the ad
    img2x3 : str
        Link to get the 2x3 image of the ad
    img16x9 : str
        Link to get the 16x9 image of the ad
    img64x10 : str
        Link to get the 64x10 image of the ad
    media : str
        Link to get the media of the ad
    display_format : str
        The display format of the ad
    """

    def __init__(
        self,
        uid: str,
        name: str,
        type: str,
        url: str,
        img2x3: str,
        img16x9: str,
        img64x10: str,
        media: str,
        display_format: str,
    ):
        self.uid = uid
        self.name = name
        self.type = type
        self.url = url
        self.img2x3 = img2x3
        self.img16x9 = img16x9
        self.img64x10 = img64x10
        self.media = media
        self.display_format = display_format

    @classmethod
    def _from_dict(cls: Self, raw: Dict) -> Self:
        uid = raw.get("uid")
        name = raw.get("name")
        type = raw.get("type")
        url = raw.get("url")
        img2x3 = raw.get("img2x3")
        img16x9 = raw.get("img16x9")
        img64x10 = raw.get("img64x10")
        media = raw.get("media")
        display_format = raw.get("displayformat")

        args = [uid, name, type, url, img2x3, img16x9, img64x10, media, display_format]
        return cls(*args)

    @classmethod
    async def list_ads(cls) -> List[Self]:
        """
        .. versionadded :: 0.3.0
        .. versionupdated :: 0.4.0
            Changed to classmethod

        Lists all ads currently in trackmania.

        Returns
        -------
        :class:`List[Self]`
            All the Ads

        Raises
        ------
        :class:`TMIOException`
            If an unexpected error occurs.
        """
        _log.info("Listing all Ads")
        try:
            all_ads = await _get_ad_list()
        except TMIOException as excp:
            raise TMIOException("An Unexpected Error Occured") from excp

        ad_list = []
        for ad in all_ads:
            ad_list.append(cls._from_dict(ad))
        return ad_list

    @classmethod
    async def get_ad(cls, ad_uid: str) -> Self | None:
        """
        .. versionadded :: 0.3.0
        .. versionupdated :: 0.4.0
            Changed to classmethod

        Gets an ad by its unique ID.

        Parameters
        ----------
        ad_uid : str
            The unique ID of the ad

        Returns
        -------
        :class:`Self`
            The ad
        """
        _log.info("Getting Ad with UID: %s", ad_uid)
        all_ads = await cls.list_ads()
        for ad in all_ads:
            if ad.uid == ad_uid:
                _log.debug("Ad with UID %s found", ad_uid)
                return ad

        _log.debug("No Ad With UID %s found", ad_uid)
        return None



import json
import logging
from contextlib import suppress
from datetime import datetime
from types import NoneType
from typing import Dict, List

import redis
from typing_extensions import Self

from ._util import _add_commas
from .api import _APIClient
from .config import Client
from .constants import _TMIO
from .errors import InvalidIDError, InvalidTrophyNumber, TMIOException

_log = logging.getLogger(__name__)

__all__ = ("PlayerTrophies",)


class TrophyLeaderboardPlayer:
    """
    .. versionadded :: 0.4.0

    Represents a player on the trophy leaderboards

    Parameters
    ----------
    player_name : str
        The player's name
    club_tag : str | None
        The player's club tag
    player_id : str
        The player's ID
    rank : int
        The player's rank
    score : str
        The player's score
    zones : `List[PlayerZone]`
        The player's zones
    """

    def __init__(
        self,
        player_name: str,
        club_tag: str | None,
        player_id: str,
        rank: int,
        score: str,
        zones: List,
    ):
        self.player_name = player_name
        self.club_tag = club_tag
        self.player_id = player_id
        self.rank = rank
        self.score = score
        self.zones = zones

    @classmethod
    def _from_dict(cls: Self, raw: Dict) -> Self:
        from .player import PlayerZone

        args = []
        player = raw.get("player")
        args.append(player.get("name"))
        args.append(player.get("tag", None))
        args.append(player.get("id"))
        args.append(raw.get("rank"))
        args.append(_add_commas(int(raw.get("score"))))
        args.append(PlayerZone._parse_zones(player.get("zone"), [0, 0, 0, 0, 0]))

        return cls(*args)

    async def get_player(self: Self):
        """
        .. versionadded :: 0.4.0

        Gets the player object using the player ID.

        Returns
        -------
        :class:`Player`
            The player object
        """
        from .player import Player

        return await Player.get_player(self.player_id)


class PlayerTrophies:
    """
    .. versionadded :: 0.1.0

    Represents Player Trophies

    Parameters
    ----------
    echelon : int
        The trophy echelon of the player.
    last_change : str
        The date of the last change of the player's self.
    points : ints: int
        The number of points of the player.
    trophies : :class:`List[int]`
        The number of trophies of the player.
    player_id : str | :class:`NoneType`, optional
        The Trackmania ID of the player
    """

    def __init__(
        self,
        echelon: int,
        last_change: datetime,
        points: int,
        trophies: List[int],
        player_id: str | NoneType = None,
    ):
        """Constructor for the class."""
        self.echelon = echelon
        self._last_change = last_change
        self.points = points
        self.trophies = trophies
        self._player_id = player_id

    @classmethod
    def _from_dict(cls: Self, raw_trophy_data: Dict, player_id: str) -> Self:
        """
        Creates a :class:`PlayerTrophies` object from the given dictionary.

        Parameters
        ----------
        raw_trophy_data : :class:`Dict`
            The raw trophy data to parse.
        player_id : str
            The player ID to set.
        Returns
        -------
        :class:`PlayerTrophies`
            The parsed trophy data.
        """
        _log.debug("Creating a PlayerTrophies class from the given dictionary.")

        return cls(
            echelon=raw_trophy_data.get("echelon"),
            last_change=datetime.strptime(
                raw_trophy_data.get("timestamp"), "%Y-%m-%dT%H:%M:%S+00:00"
            ),
            points=raw_trophy_data.get("points"),
            trophies=raw_trophy_data.get("counts"),
            player_id=player_id,
        )

    @property
    def last_change(self):
        """Last change property."""
        return self._last_change

    @property
    def player_id(self):
        """player_id property"""
        return self._player_id

    def set_id(self, player_id: str):
        """Setter for player_id"""
        self._player_id = player_id

    def trophy(self, number: int) -> int:
        """
        .. versionadded :: 0.3.0

        Returns the trophies by tier.

        Parameters
        ----------
        number : int
            The trophy number, from 1 (T1) to 9 (T9).
        Returns
        -------
        int
            the number of trophies for that specific tier.
        """
        _log.debug(f"Returning trophy T{number} for player {self.player_id}")

        if number > 9 or number < 1:
            raise InvalidTrophyNumber(
                "Trophy Number cannot be less than 1 or greater than 9"
            )

        return self.trophies[number - 1]

    def score(self) -> int:
        """Returns the trophy score of the player."""

        score = (
            0
            + self.trophy(1) * 1
            + self.trophy(2) * 10
            + self.trophy(3) * 100
            + self.trophy(4) * 1_000
            + self.trophy(5) * 10_000
            + self.trophy(6) * 100_000
            + self.trophy(7) * 1_000_000
            + self.trophy(8) * 10_000_000
        )

        _log.debug(f"Score of {self.player_id} is {score}")

        return score

    def __str__(self) -> str:
        trophy_str = ""
        for i, trophyd in enumerate(self.trophies):
            trophy_str = trophy_str + f"T{i + 1} - " + _add_commas(trophyd) + "\n"

        trophy_str = trophy_str + f"\nTotal Trophies: {_add_commas(self.score())}"

        return trophy_str

    async def history(self, page: int = 0) -> Dict:
        """
        .. versionadded :: 0.3.0

        Retrieves Trophy Gain and Loss history of a player.

        Parameters
        ----------
        page : int, optional
            page number of trophy history, by default 0
        Returns
        -------
        :class:`Dict`
            Trophy history data.
        Raises
        ------
        InvalidIDError
            If an ID has not been set for the object.
        InvalidIDError
            If an invalid id has been set for the object.
        """
        _log.debug(
            f"Getting Trophy Leaderboard for Page: {page} and Player Id: {self.player_id}"
        )

        cache_client = Client._get_cache_client()

        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            if cache_client.exists(f"trophy:{page}"):
                _log.debug(f"Found trophy leaderboard for page {page} in cache")
                return json.loads(cache_client.get(f"trophy:{page}").decode("utf-8"))[
                    "gains"
                ]

        api_client = _APIClient()

        if self.player_id is None:
            raise InvalidIDError("ID Has not been set for the Object")

        history = await api_client.get(
            _TMIO.build(
                [_TMIO.TABS.PLAYER, self.player_id, _TMIO.TABS.TROPHIES, str(page)]
            )
        )

        await api_client.close()

        with suppress(KeyError, TypeError):
            raise TMIOException(history["error"])
        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            _log.debug(f"Caching trophy history for page: {page}")
            cache_client.set(f"trophy:{page}", json.dumps(history), ex=3600)

        return history["gains"]

    @staticmethod
    async def top(page: int = 0) -> List[TrophyLeaderboardPlayer]:
        """
        .. versionadded :: 0.3.0

        Get's the top players ranked by trophies

        Parameters
        ----------
        page : int, optional
            The page of the leaderboards, by default 0

        Returns
        -------
        :class:`List[TrophyLeaderboardPlayer]`
            The players as a list of :class:`TrophyLeaderboardPlayer` objects.
        """
        _log.debug(f"Getting Page {page} of Trophy Leaderboards")

        cache_client = Client._get_cache_client()

        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            if cache_client.exists(f"trophies:{page}"):
                _log.debug(f"Found trophy leaderboard for page {page} in cache")
                all_players = json.loads(
                    cache_client.get(f"trophies:{page}").decode("utf-8")
                )

                lb_players = []
                for top_player in all_players["ranks"]:
                    lb_players.append(TrophyLeaderboardPlayer._from_dict(top_player))

        api_client = _APIClient()

        top_trophies = await api_client.get(
            _TMIO.build([_TMIO.TABS.TOP_TROPHIES, str(page)])
        )

        await api_client.close()

        with suppress(KeyError, TypeError):
            raise TMIOException(top_trophies["error"])
        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            _log.debug(f"Caching trophy leaderboard for page: {page}")
            cache_client.set(f"trophies:{page}", json.dumps(top_trophies), ex=3600)

        lb_players = []
        for top_player in top_trophies["ranks"]:
            lb_players.append(TrophyLeaderboardPlayer._from_dict(top_player))

        return lb_players



import json
import logging
from contextlib import suppress
from datetime import datetime
from typing import Dict, List

import redis
from typing_extensions import Self

from trackmania.api import _APIClient

from .api import _APIClient
from .config import Client
from .constants import _TMX
from .errors import InvalidTMXCode

_log = logging.getLogger(__name__)

__all__ = (
    "TMXMapTimes",
    "GbxFileMetadata",
    "TMXTags",
    "ReplayWRData",
    "TMXMetadata",
    "TMXMap",
)


async def _get_map(tmx_id: int) -> Dict:
    _log.info(f"Getting map data for tmx id {tmx_id}")

    cache_client = Client._get_cache_client()
    with suppress(ConnectionError, redis.exceptions.ConnectionError):
        if cache_client.exists(f"tmxmap:{tmx_id}"):
            _log.debug(f"Found map data for tmx id {tmx_id} in cache")
            return json.loads(cache_client.get(f"tmxmap:{tmx_id}").decode("utf-8"))

    api_client = _APIClient()
    map_data = await api_client.get(
        _TMX.build([_TMX.TABS.MAPS, _TMX.TABS.GET_MAP_INFO, _TMX.TABS.ID, tmx_id])
    )
    await api_client.close()

    if not isinstance(map_data, dict):
        raise InvalidTMXCode("Invalid TMX code")
    with suppress(ConnectionError, redis.exceptions.ConnectionError):
        _log.debug(f"Caching map data for tmx id {tmx_id}")
        cache_client.set(f"tmxmap:{tmx_id}", json.dumps(map_data))

    return map_data


class TMXMapTimes:
    """
    .. versionadded :: 0.3.3

    Represents when the map was uploaded and/or updated on the `trackmania.exchange` website

    Parameters
    ----------
    uploaded : :class:`datetime`
        When the map was uploaded to TMX
    updated : :class:`datetime`
        When the map was last updated on TMX
    """

    def __init__(self, uploaded: datetime, updated: datetime):
        self.uploaded = uploaded
        self.updated = updated

    @classmethod
    def _from_dict(cls: Self, raw: Dict) -> Self:
        _log.debug("Creating a TMXMapTimes from given dictionary")

        uploaded_raw, updated_raw = raw.get("UploadedAt"), raw.get("UpdatedAt")

        # 2022-03-15T18:18:50.007 to datetime
        uploaded = datetime.strptime(uploaded_raw, "%Y-%m-%dT%H:%M:%S.%f")
        updated = datetime.strptime(updated_raw, "%Y-%m-%dT%H:%M:%S.%f")

        return cls(uploaded, updated)


class GbxFileMetadata:
    """
    .. versionadded :: 0.3.3

    Represents data stored in the map's GBX file metadata

    Parameters
    ----------
    gbx_map_name : str
        Unaltered map name
    author_login : str
        Unaltered author name
    map_type : str
        Nadeo map type embedded.
    title_pack : str
        Titlepack of map
    track_uid : str
        Unaltered map signature `uid`
    mood : str
        Mood of the map (daytime and map base)
    display_cost : int
        in game metric for map weight
    mod_name : str | None
        Name of texture mod, if included
    light_map : int
        Light map version
    exe_version : str
        Game application version
    exe_build : :class:`datetime`
        Build date of game version
    author_time : int
        Map author time in ms
    environment_name : str
        Name of the game environment
    vehicle_name : str
        Name of the game vehicle
    """

    def __init__(
        self,
        gbx_map_name: str,
        author_login: str,
        map_type: str,
        title_pack: str,
        track_uid: str,
        mood: str,
        display_cost: int,
        mod_name: str,
        light_map: int,
        exe_version: str,
        exe_build: datetime,
        author_time: int,
        environment_name: str,
        vehicle_name: str,
    ):
        self.gbx_map_name = gbx_map_name
        self.author_login = author_login
        self.map_type = map_type
        self.title_pack = title_pack
        self.track_uid = track_uid
        self.mood = mood
        self.display_cost = display_cost
        self.mod_name = mod_name
        self.light_map = light_map
        self.exe_version = exe_version
        self.exe_build = exe_build
        self.author_time = author_time
        self.environment_name = environment_name
        self.vehicle_name = vehicle_name

    @classmethod
    def _from_dict(cls: Self, raw: Dict) -> Self:
        _log.debug("Creating a GbxFileMetadata from given dictionary")

        args = [
            raw.get("GbxMapName"),
            raw.get("AuthorLogin"),
            raw.get("MapType"),
            raw.get("TitlePack"),
            raw.get("TrackUID"),
            raw.get("Mood"),
            raw.get("DisplayCost"),
            raw.get("ModName"),
            raw.get("Lightmap"),
            raw.get("ExeVersion"),
            datetime.strptime(raw.get("ExeBuild"), "%Y-%m-%d_%H_%M"),
            raw.get("AuthorTime"),
            raw.get("EnvironmentName"),
            raw.get("VehicleName"),
        ]

        return cls(*args)


class TMXTags:
    """
    .. versionadded :: 0.3.3

    Represents TMX Map Tags

    Parameters
    ----------
    map_tags : :class:`List[int]`
        The map tags as a list of integers
    """

    def __init__(self, map_tags: List[int]):
        self.map_tags = map_tags

    @classmethod
    def _from_dict(cls: Self, raw: Dict) -> Self:
        _log.debug("Creating a TMXTags from given dictionary")

        map_tags_ss = raw.get("Tags", None).split(",")
        map_tags = []

        for tag in map_tags_ss:
            map_tags.append(int(tag))

        return cls(map_tags)

    def __str__(self) -> str:
        return ", ".join([_TMX.MAP_TYPE_ENUMS.get(tag) for tag in self.map_tags])


class ReplayWRData:
    """
    .. versionadded :: 0.3.3

    Represents data relating to the wr replay

    Parameters
    ---------
    wr_id : int | None
        replay id of mx world record
    wr_time: int | None
        Time of the mx world record
    wr_user_id: int | None
        userid of the mx world record holder
    wr_username: str
        mx username of max world record holder
    """

    def __init__(
        self,
        wr_id: int | None,
        wr_time: int | None,
        wr_user_id: int | None,
        wr_username: str,
    ):
        self.wr_id = wr_id
        self.wr_time = wr_time
        self.wr_user_id = wr_user_id
        self.wr_username = wr_username

    @classmethod
    def _from_dict(cls: Self, raw: Dict) -> Self:
        _log.debug("Creating a ReplayWRData from given dictionary")

        args = [
            raw.get("ReplayWRId"),
            raw.get("ReplayWRTime"),
            raw.get("ReplayWRUserID"),
            raw.get("ReplayWRUsername"),
        ]

        return cls(*args)


class TMXMetadata:
    """
    .. versionadded :: 0.3.3

    Represents a map's metadata on TMX

    Parameters
    ----------
    unlisted : bool
        Whether the map is unlisted on TMX
    unreleased : bool
        Whether the map is unreleased on TMX
    downloadeable : bool
        Whether the map is downloadeable on TMX
    rating_vote_count : int
        The number of votes for the map
    rating_vote_average : float
        The average rating of the map
    has_screenshot : bool
        Whether the map has a screenshot
    has_thumbnail : bool
        Whether the map has a thumbnail
    has_ghost_blocks : bool
        Whether the map has ghost blocks
    embedded_objects_count : int
        The number of embedded objects on the map
    embedded_objects_size : int
        The size of the embedded objects on the map
    size_warning : bool
        Map exceeds the size limit for online servers (TM: 6MB)
    replay_count : int
        The number of replays on the map
    award_count : int
        The number of awards on the map
    comments_count : int
        The number of comments on the map
    image_count : int
        The number of custom images on the map
    video_count : int
        The number of videos on the map
    """

    def __init__(
        self,
        unlisted: bool,
        unreleased: bool,
        downloadable: bool,
        rating_vote_count: int,
        rating_vote_average: float,
        has_screenshot: bool,
        has_thumbnail: bool,
        has_ghost_blocks: bool,
        embedded_objects_count: int,
        embedded_objects_size: int,
        size_warning: bool,
        replay_count: int,
        award_count: int,
        comment_count: int,
        image_count: int,
        video_count: int,
    ):
        self.unlisted = unlisted
        self.unreleased = unreleased
        self.downloadable = downloadable
        self.rating_vote_count = rating_vote_count
        self.rating_vote_average = rating_vote_average
        self.has_screenshot = has_screenshot
        self.has_thumbnail = has_thumbnail
        self.has_ghost_blocks = has_ghost_blocks
        self.embedded_objects_count = embedded_objects_count
        self.embedded_objects_size = embedded_objects_size
        self.size_warning = size_warning
        self.replay_count = replay_count
        self.award_count = award_count
        self.comment_count = comment_count
        self.image_count = image_count
        self.video_count = video_count

    @classmethod
    def _from_dict(cls: Self, raw: Dict) -> Self:
        _log.debug("Creating a TMXMetaData from given dictionary")

        args = [
            raw.get("Unlisted"),
            raw.get("Unreleased"),
            raw.get("Downloadable"),
            raw.get("RatingVoteCount"),
            raw.get("RatingVoteAverage"),
            raw.get("HasScreenshot"),
            raw.get("HasThumbnail"),
            raw.get("HasGhostBlocks"),
            raw.get("EmbeddedObjectsCount"),
            raw.get("EmbeddedItemsSize"),
            raw.get("SizeWarning"),
            raw.get("ReplayCount"),
            raw.get("AwardCount"),
            raw.get("CommentCount"),
            raw.get("ImageCount"),
            raw.get("VideoCount"),
        ]

        return cls(*args)


class TMXMap:
    """
    .. versionadded :: 0.3.3

    Represents a Map from TMX

    Parameters
    ----------
    username : str
        The username of the map's creator
    track_id : int | None
        The track id of the map
    map_id : int | None
        The map id of the map
    comments : str
        Map author's comments
    map_pack_id : str
        The map pack id of the map
    user_id : int
        The user id of the map's creator
    route_name : str
        The route name of the map
    length_name : str
        The length name of the map
    difficulty_name : str
        The difficulty name of the map
    laps : int
        The number of laps of the map
    gbx_data : :class:`GbxFileMetadata`
        The gbx file metadata of the map
    tags : :class:`TMXTags`
        The tags of the map
    replay_wr_data : :class:`ReplayWRData`
        The replay wr data of the map
    metadata : :class:`TMXMetadata`
        The metadata of the map
    """

    def __init__(
        self,
        username: str,
        track_id: int | None,
        map_id: int | None,
        comments: str,
        map_pack_id: str,
        user_id: int,
        route_name: str,
        length_name: str,
        difficulty_name: str,
        laps: int,
        gbx_data: GbxFileMetadata,
        tags: TMXTags,
        replay_wr_data: ReplayWRData,
        metadata: TMXMetadata,
    ):
        self.username = username
        self.track_id = track_id
        self.map_id = map_id
        self.comments = comments
        self.map_pack_id = map_pack_id
        self.user_id = user_id
        self.route_name = route_name
        self.length_name = length_name
        self.difficulty_name = difficulty_name
        self.laps = laps
        self.gbx_data = gbx_data
        self.tags = tags
        self.replay_wr_data = replay_wr_data
        self.metadata = metadata

    @classmethod
    def _from_dict(cls: Self, raw: Dict) -> Self:
        _log.debug("Creating a TMXMap from given dictionary")

        username = raw.get("Username")
        track_id = raw.get("TrackID")
        map_id = raw.get("MapID")
        comments = raw.get("Comments")
        map_pack_id = raw.get("MappackID")
        user_id = raw.get("UserID")
        route_name = raw.get("RouteName")
        length_name = raw.get("LengthName")
        difficulty_name = raw.get("DifficultyName")
        laps = raw.get("Laps")
        gbx_data = GbxFileMetadata._from_dict(raw)
        tags = TMXTags._from_dict(raw)
        replay_wr_data = ReplayWRData._from_dict(raw)
        metadata = TMXMetadata._from_dict(raw)

        args = [
            username,
            track_id,
            map_id,
            comments,
            map_pack_id,
            user_id,
            route_name,
            length_name,
            difficulty_name,
            laps,
            gbx_data,
            tags,
            replay_wr_data,
            metadata,
        ]

        return cls(*args)

    @classmethod
    async def get_map(cls: Self, tmx_id: int) -> Self:
        """
        .. versionadded :: 0.3.3

        Gets a map's data using it's tmx id

        Parameters
        ----------
        tmx_id : int
            The tmx id
        """
        tmx_map_data = await _get_map(tmx_id)
        return cls._from_dict(tmx_map_data)



__all__ = (
    "TrackmaniaException",
    "TMIOException",
    "InvalidPlayerException",
    "NoUserAgentSetError",
    "InvalidUsernameError",
    "InvalidIDError",
    "InvalidTrophyNumber",
    "InvalidTOTDDate",
)


# pylint: disable=unnecessary-pass
class TrackmaniaException(Exception):
    """BASE exception class for py-tmio"""

    pass


class TMIOException(Exception):
    """BASE exception class for errors from trackmania.io"""

    pass


class TMXException(Exception):
    """BASE exception class for errors from trackmania.exchange"""

    pass


class InvalidPlayerException(TrackmaniaException):
    """Base Exception class for Player-Related exceptions"""

    pass


class NoUserAgentSetError(Exception):
    """Raised when a User-Agent is not set."""

    def __init__(self):
        message = "No User Agent has been set.\nPlease read the README for instructions on how to set the USER_AGENT."
        super().__init__(message)


class InvalidUsernameError(InvalidPlayerException):
    """Raised when a username is not valid."""

    def __init__(self, *args):
        if args:
            message = args[0]
        else:
            message = None

        super().__init__(message)


class InvalidIDError(InvalidPlayerException):
    """Raised when an Invalid ID is given."""

    def __init__(self, *args):
        if args:
            message = args[0]
        else:
            message = None

        super().__init__(message)


class InvalidTrophyNumber(TrackmaniaException):
    """ """

    def __init__(self, *args):
        if args:
            message = args[0]
        else:
            message = None

        super().__init__(message)


class InvalidTOTDDate(TrackmaniaException):
    """Raised when an invalid TOTD Date is given."""

    def __init__(self, *args):
        if args:
            message = args[0]
        else:
            message = None

        super().__init__(message)


class InvalidTMXCode(TMXException):
    """Raised when an invalid TMX Code is given."""

    def __init__(self, *args):
        if args:
            message = args[0]
        else:
            message = None

        super().__init__(message)



"""
Trackmania.io API Wrapper
~~~~~~~~~~~~~~~~~~~~~~~~~

A wrapper for Trackmania.io api service.

:copyright: (c) 2022-present Deepesh Nimma.
:license: MIT, see LICENSE for more details.
"""
import asyncio
import logging
import os

from .ad import *
from .config import *
from .cotd import *
from .errors import *
from .matchmaking import *
from .player import *
from .tmmap import *
from .tmx import *
from .totd import *
from .trophy import *

__title__ = "py-tmio"
__author__ = "Deepesh Nimma"
__license__ = "MIT"
__copyright__ = "Copyright 2022-present Deepesh Nimma"
__version__ = "v0.4.0b1"

__path__ = __import__("pkgutil").extend_path(__path__, __name__)

if os.name == "nt":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

logging.getLogger(__name__).addHandler(logging.NullHandler())



import json
import logging
from contextlib import suppress
from datetime import datetime
from typing import Dict, List

import redis
from typing_extensions import Self

from trackmania.api import _APIClient

from .api import _APIClient
from .config import Client
from .constants import _TMIO
from .errors import TMIOException
from .player import Player

_log = logging.getLogger(__name__)

__all__ = (
    "MedalTimes",
    "Leaderboard",
    "TMMap",
)


class MedalTimes:
    """
    .. versionadded :: 0.3.0

    Represents a map's medal times

    Parameters
    ----------
    bronze : int
        The bronze medal time in ms
    silver : int
        The silver medal time in ms
    gold : int
        The gold medal time in ms
    author : int
        The author of the medal times
    bronze_string : str
        The bronze medal time in mm:ss:msmsms format
    silver_string : str
        The silver medal time in mm:ss:msmsms format
    gold_string : str
        The gold medal time in mm:ss:msmsms format
    author_string : str
        The author of the medal times in mm:ss:msmsms format
    """

    def __init__(self, bronze: int, silver: int, gold: int, author: int):
        self.bronze = bronze
        self.silver = silver
        self.gold = gold
        self.author = author
        self.bronze_string = self._parse_to_string(self.bronze)
        self.silver_string = self._parse_to_string(self.silver)
        self.gold_string = self._parse_to_string(self.gold)
        self.author_string = self._parse_to_string(self.author)

    def _parse_to_string(self, time: int) -> str:
        """
        Parses a medal time to a string in format `mm:ss:msms`

        Parameters
        ----------
        time : int
            The time to parse

        Returns
        -------
        str
            The time in mm:ss:msmsms format
        """
        _log.debug(f"Parsing {time} to string")

        sec, ms = divmod(time, 1000)
        min, sec = divmod(time, 60)

        return f"{min:02d}:{sec:02d}.{ms:03d}"


class Leaderboard:
    """
    .. versionadded :: 0.3.0

    Represents the Map's Leaderboard

    Parameters
    ----------
    timestamp : :class:`datetime`
        The timestamp of the leaderboard was achieved
    ghost : str
        The url for the ghost download
    player_club_tag : str | None
        The player's club tag
    player_name : str
        The player's name
    player_id : str
        .. versionadded :: 0.3.4
        The player's id
    position : int
        The position of the player in the leaderboard
    time : int
        The time of the player in the leaderboard
    """

    def __init__(
        self,
        timestamp: datetime,
        ghost: str,
        player_club_tag: str | None,
        player_name: str | None,
        player_id: str | None,
        position: int,
        time: int,
    ):
        self.timestamp = timestamp
        self.ghost = ghost
        self.player_club_tag = player_club_tag
        self.player_name = player_name
        self.position = position
        self.time = time
        self.player_id = player_id

    @classmethod
    def _from_dict(cls: Self, raw: Dict) -> Self:
        _log.debug("Creating a Leaderboards class from given dictionary")

        if "player" in raw:
            player_id = raw.get("player").get("id")
            player_name = raw.get("player").get("name")
            player_club_tag = raw.get("player").get("tag", None)
        else:
            player_id = None
            player_name = None
            player_club_tag = None

        position = raw.get("position")
        time = raw.get("time")
        ghost = raw.get("url")
        timestamp = datetime.strptime(raw.get("timestamp"), "%Y-%m-%dT%H:%M:%S+00:00")

        return cls(
            timestamp=timestamp,
            ghost=ghost,
            player_club_tag=player_club_tag,
            player_name=player_name,
            position=position,
            time=time,
            player_id=player_id,
        )

    async def get_player(self) -> Player:
        """
        .. versionadded :: 0.3.4

        Gets the player who achieved the leaderboard.

        Returns
        -------
        :class:`Player`
            The player obj, None if the player_id does not exist.
        """
        if self.player_id is None:
            return None

        return await Player.get_player(self.player_id)


class TMMap:
    """
    .. versionadded :: 0.3.0

    Represents a Trackmania Map

    Parameters
    ----------
    author_id : str
        The author's player id
    environment : str
        The environment of the map
    exchange_id : str | None
        The exchange id of the map
    file_name : str
        The file name of the map
    map_id : str
        The map id of the map
    leaderboard : :class:`List[Leaderboard]` | None
        The leaderboard of the map
    medal_times : :class:`MedalTimes`
        The medal times of the map
    name : str
        The name of the map
    submitter_id : str
        The map's submitter's id
    thumbnail : str
        Link to the thumbnail of the map
    uid : str
        The uid of the map
    uploaded : :class:`datetime`
        The timestamp of when the map was uploaded
    url : str
        The url of the map download
    lb_loaded : bool
        Whether the leaderboard has been loaded
    """

    def __init__(
        self,
        author_id: str,
        author_name: str,
        environment: str,
        exchange_id: str | None,
        file_name: str,
        map_id: str,
        leaderboard: List[Leaderboard] | None,
        medal_time: MedalTimes,
        name: str,
        submitter_id: str,
        submitter_name: str,
        thumbnail: str,
        uid: str,
        uploaded: datetime,
        url: str,
    ):
        self.author_id = author_id
        self.author_name = author_name
        self.environment = environment
        self.exchange_id = exchange_id
        self.file_name = file_name
        self.map_id = map_id
        self.leaderboard = leaderboard
        self.medal_time = medal_time
        self.name = name
        self.submitter_id = submitter_id
        self.submitter_name = submitter_name
        self.thumbnail = thumbnail
        self.uid = uid
        self.uploaded = uploaded
        self.url = url
        self._offset = 0
        self.length = 100
        self._lb_loaded = False

    @property
    def offset(self):
        return self._offset

    @property
    def lb_loaded(self):
        return self._lb_loaded

    @classmethod
    def _from_dict(cls: Self, raw: Dict) -> Self:
        _log.debug("Creating a Map class from given dictionary")

        author_id = raw.get("author")
        author_name = raw.get("authorplayer").get("name")
        environment = raw.get("collectionName")
        exchange_id = raw.get("exchangeid", None)
        file_name = raw.get("filename")
        map_id = raw.get("mapId")
        leaderboard = None
        medal_time = MedalTimes(
            raw.get("bronzeScore"),
            raw.get("silverScore"),
            raw.get("goldScore"),
            raw.get("authorScore"),
        )
        name = raw.get("name")
        submitter_id = raw.get("submitter")
        submitter_name = raw.get("submitterplayer").get("name")
        thumbnail = raw.get("thumbnailUrl")
        uid = raw.get("mapUid")
        uploaded = datetime.strptime(raw.get("timestamp"), "%Y-%m-%dT%H:%M:%S+00:00")
        url = raw.get("fileUrl")

        return cls(
            author_id,
            author_name,
            environment,
            exchange_id,
            file_name,
            map_id,
            leaderboard,
            medal_time,
            name,
            submitter_id,
            submitter_name,
            thumbnail,
            uid,
            uploaded,
            url,
        )

    @classmethod
    async def get_map(cls: Self, map_uid: str) -> Self:
        """
        .. versionadded :: 0.3.0

        Gets the TM Map from the Map's UID

        Parameters
        ----------
        map_uid : str
            The map's UID
        """
        _log.debug(f"Getting the map with the UID {map_uid}")

        cache_client = Client._get_cache_client()

        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            if cache_client.exists(f"map:{map_uid}"):
                _log.debug(f"Map {map_uid} found in cache")
                return cls._from_dict(json.loads(cache_client.get(f"map:{map_uid}")))

        api_client = _APIClient()
        map_data = await api_client.get(_TMIO.build([_TMIO.TABS.MAP, map_uid]))
        await api_client.close()

        with suppress(KeyError, TypeError):

            raise TMIOException(map_data["error"])
        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            _log.debug(f"Caching map {map_uid}")
            cache_client.set(f"map:{map_uid}", json.dumps(map_data))

        return cls._from_dict(map_data)

    async def author(self) -> Player:
        """
        .. versionadded :: 0.3.0

        Returns the author as a player.

        Returns
        -------
        :class:`Player`
            The author as a :class:`Player` object
        """
        _log.debug(f"Getting the author of the map {self.uid}")
        return await Player.get_player(self.author_id)

    async def submitter(self) -> Player:
        """
        .. versionadded :: 0.3.0

        Returns the submitter as a player.

        Returns
        -------
        :class:`Player`
            The submitter as a :class:`Player` object
        """
        _log.debug(f"Getting the submitter of the map {self.uid}")
        return await Player.get_player(self.submitter_id)

    async def get_leaderboard(
        self, offset: int = 0, length: int = 100
    ) -> List[Leaderboard]:
        """
        .. versionadded :: 0.3.0

        Get's the leaderboard of a map.

        Parameters
        ----------
        offset : int, optional
            The offset of the leaderboard. Defaults to 0.
        length : int, optional
            How many leaderpositions to get. Should be between 1 and 100 both inclusive. by default 100

        Returns
        -------
        :class:`List[Leaderboard]`
            The leaderboard as a list of positions.

        Raises
        ------
        :class:`ValueError`
            If the length is not between 1 and 100.
        """
        if length < 1:
            raise ValueError("Length must be greater than 0")
        length = min(length, 100)

        _log.debug(
            f"Getting Leaderboard of the Map with Length {length} and offset {offset}"
        )

        cache_client = Client._get_cache_client()

        self._offset = offset
        self.length = length

        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            if cache_client.exists(
                f"leaderboard:{self.uid}:{self.offset}:{self.length}"
            ):
                _log.debug(
                    f"Leaderboard {self.uid}:{self.offset}:{self.length} found in cache"
                )
                leaderboards = []
                for lb in json.loads(
                    cache_client.get(
                        f"leaderboard:{self.uid}:{self.offset}:{self.length}"
                    ).decode("utf-8")
                )["tops"]:
                    leaderboards.append(Leaderboard._from_dict(lb))

                return leaderboards

        api_client = _APIClient()
        lb_data = await api_client.get(
            _TMIO.build([_TMIO.TABS.LEADERBOARD, _TMIO.TABS.MAP, self.uid])
            + f"?offset={self.offset}&length={self.length}"
        )
        await api_client.close()

        with suppress(KeyError, TypeError):

            raise TMIOException(lb_data["error"])
        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            _log.debug(f"Caching leaderboard {self.uid}:{self.offset}:{self.length}")
            cache_client.set(
                f"leaderboard:{self.uid}:{self.offset}:{self.length}",
                json.dumps(lb_data),
            )

        self._offset += self.length
        self._lb_loaded = True

        leaderboards = []
        for lb in lb_data["tops"]:
            leaderboards.append(Leaderboard._from_dict(lb))

        return leaderboards

    async def load_more_leaderboard(self, length: int = 100) -> List[Leaderboard]:
        """
        .. versionadded :: 0.3.0

        Gets more leaderboards for the map. If `get_leaderboards` wasn't used before then it just gets it from the start.

        Parameters
        ----------
        length : int, optional
            How many leaderboard positions to get, by default 100

        Returns
        -------
        :class:`List[Leaderboard]`
            The leaderboard positions.
        """
        cache_client = Client._get_cache_client()

        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            if cache_client.exists(
                f"leaderboard:{self.uid}:{self.offset}:{self.length}"
            ):
                _log.debug(
                    f"Leaderboard {self.uid}:{self.offset}:{self.length} found in cache"
                )
                leaderboards = []
                for lb in json.loads(
                    cache_client.get(
                        f"leaderboard:{self.uid}:{self.offset}:{self.length}"
                    ).decode("utf-8")
                )["tops"]:
                    leaderboards.append(Leaderboard._from_dict(lb))

                return leaderboards

        if not self._lb_loaded:
            _log.warn("Leaderboard is not loaded yet, loading from start")
            return await self.get_leaderboard(length=length)

        api_client = _APIClient()
        leaderboards = await api_client.get(
            _TMIO.build([_TMIO.TABS.LEADERBOARD, _TMIO.TABS.MAP, self.uid])
            + f"?offset={self._offset}&length={length}"
        )
        await api_client.close()

        with suppress(KeyError, TypeError):
            raise TMIOException(leaderboards["error"])
        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            _log.debug(f"Caching leaderboard {self.uid}:{self.offset}:{self.length}")
            cache_client.set(
                f"leaderboard:{self.uid}:{self.offset}:{self.length}",
                json.dumps(leaderboards),
            )

        self._offset += length
        self._lb_loaded = True

        lbs = []
        for lb in leaderboards["tops"]:
            lbs.append(Leaderboard._from_dict(lb))

        return lbs



import json
import logging
from contextlib import suppress
from datetime import datetime
from types import NoneType
from typing import Dict, List

import redis
from typing_extensions import Self

from trackmania.errors import TMIOException

from .api import _APIClient
from .config import Client
from .constants import _TMIO
from .errors import InvalidIDError, TMIOException

_log = logging.getLogger(__name__)

__all__ = (
    "BestCOTDStats",
    "PlayerCOTDStats",
    "PlayerCOTDResults",
    "PlayerCOTD",
    "COTD",
)


async def _get_trophy_page(player_id: str, page: int) -> Dict:
    _log.debug(f"Getting COTD Stats for Player {player_id} and page {page}")

    cache_client = Client._get_cache_client()

    with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
        if cache_client.exists(f"playercotd:{player_id}:{page}"):
            _log.debug(f"Player {player_id}'s page {page} cotd results found in cache")
            return json.loads(
                cache_client.get(f"playercotd:{player_id}:{page}").decode("utf-8")
            )

    api_client = _APIClient()
    page_data = await api_client.get(
        _TMIO.build([_TMIO.TABS.PLAYER, player_id, _TMIO.TABS.COTD, str(page)])
    )
    await api_client.close()

    with suppress(KeyError, TypeError):
        raise TMIOException(page_data["error"])
    if isinstance(page_data, NoneType):
        raise InvalidIDError("Invalid PlayerID Given")
    with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
        _log.debug(f"Caching Player {player_id} Page {page}")
        cache_client.set(f"playercotd:{player_id}:{page}", json.dumps(page_data))

    return page_data


async def _get_cotd_page(page: int) -> Dict:
    _log.debug(f"Getting COTD Page {page}")

    cache_client = Client._get_cache_client()

    with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
        if cache_client.exists(f"cotd:{page}"):
            _log.debug(f"COTD Page {page} found in cache")
            all_cotds = json.loads(cache_client.get(f"cotd:{page}").decode("utf-8"))
            return all_cotds["competitions"]

    api_client = _APIClient()
    all_cotds = await api_client.get(_TMIO.build([_TMIO.TABS.COTD, page]))
    await api_client.close()

    with suppress(KeyError, TypeError):
        raise TMIOException(all_cotds["error"])
    with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
        _log.debug(f"Caching COTD Page {page}")
        cache_client.set(f"cotd:{page}", json.dumps(all_cotds), ex=7200)

    return all_cotds["competitions"]


class BestCOTDStats:
    """
    .. versionadded :: 0.3.0

    Parameters
    ----------
    best_rank : int
        The best rank achieved by the player.
    best_rank_time : datetime
        The time when `best_rank` was achieved.
    best_rank_div_rank : int
        The rank achieved in the division of the `best_rank`.
    best_div : int
        The best division of the player.
    best_div_time : datetime
        The time when `best_div` was achieved.
    best_rank_in_div : int
        The best rank the player achieved in any division.
    best_rank_in_div_time : datetime
        The time when `best_rank_in_div` was achieved.
    best_rank_in_div_div : int
        The division of the `best_rank_in_div`.
    """

    def __init__(
        self,
        best_rank: int,
        best_rank_time: datetime,
        best_rank_div_rank: int,
        best_div: int,
        best_div_time: datetime,
        best_rank_in_div: int,
        best_rank_in_div_time: datetime,
        best_rank_in_div_div: int,
    ):
        self.best_rank = best_rank
        self.best_rank_time = best_rank_time
        self.best_rank_div_rank = best_rank_div_rank
        self.best_div = best_div
        self.best_div_time = best_div_time
        self.best_rank_in_div = best_rank_in_div
        self.best_rank_in_div_time = best_rank_in_div_time
        self.best_rank_in_div_div = best_rank_in_div_div

    @classmethod
    def _from_dict(cls: Self, raw: Dict) -> Self:
        _log.debug("Creating a BestCOTDStats class from given dictionary")

        args = [
            raw.get("bestrank"),
            datetime.strptime(raw.get("bestranktime"), "%Y-%m-%dT%H:%M:%S+00:00"),
            raw.get("bestrankdivrank"),
            raw.get("bestdiv"),
            datetime.strptime(raw.get("bestdivtime"), "%Y-%m-%dT%H:%M:%S+00:00"),
            raw.get("bestrankindiv"),
            datetime.strptime(raw.get("bestrankindivtime"), "%Y-%m-%dT%H:%M:%S+00:00"),
            raw.get("bestrankindivdiv"),
        ]

        return cls(*args)


class PlayerCOTDStats:
    """
    .. versionadded :: 0.3.0

    Parameters
    ----------
    average_div : int
        The average div of the player
    average_div_rank : int
        The average div rank of the player
    average_rank : :class:`BestCOTDStats`
        The average rank of the player
    best_overall : :class:`BestCOTDStats`
        The best overall rank of the player
    best_primary : int
        The best primary rank of the player
    div_win_streak : int
        The div win streak of the player
    total_div_wins : int
        The total div wins of the player
    total_wins : int
        The total wins of the player
    win_streak : int
        The win streak of the player
    """

    def __init__(
        self,
        average_div: float,
        average_div_rank: float,
        average_rank: float,
        best_overall: BestCOTDStats,
        best_primary: BestCOTDStats,
        div_win_streak: int,
        total_div_wins: int,
        total_wins: int,
        win_streak: int,
    ):
        self.average_div = average_div
        self.average_div_rank = average_div_rank
        self.average_rank = average_rank
        self.best_overall = best_overall
        self.best_primary = best_primary
        self.div_win_streak = div_win_streak
        self.total_div_wins = total_div_wins
        self.total_wins = total_wins
        self.win_streak = win_streak

    @classmethod
    def _from_dict(cls, raw: Dict):
        _log.debug("Creating a PlayerCOTDStats class from given dictionary")

        args = [
            raw.get("avgdiv"),
            raw.get("avgdivrank"),
            raw.get("avgrank"),
            BestCOTDStats._from_dict(raw.get("bestoverall")),
            BestCOTDStats._from_dict(raw.get("bestprimary")),
            raw.get("divwinstreak"),
            raw.get("totaldivwins"),
            raw.get("totalwins"),
            raw.get("winstreak"),
        ]

        return cls(*args)


class PlayerCOTDResults:
    """
    .. versionadded :: 0.3.0

    Represents a Player's COTD Result.

    Parameters
    ----------
    id : int
        The ID of the COTD.
    timestamp : :class:`datetime`
        The timestamp of the COTD.
    name : str
        The name of the COTD
    div : int
        The division the player achieved
    rank : int
        The rank the player achieved
    div_rank : int | None
        The rank the player achieved in the division. If the player did not play in the division, this will be ``None``.
    score : int | None
        The score of the player. If the player did not play after qualifying for the COTD, this will be ``None``.
    total_players : int
        The total players that played this COTD.
    """

    def __init__(
        self,
        id: int,
        timestamp: datetime,
        name: str,
        div: int,
        rank: int,
        div_rank: int | None,
        score: int | None,
        total_players: int,
    ):
        self.id = id
        self.timestamp = timestamp
        self.name = name
        self.div = div
        self.rank = rank
        self.div_rank = div_rank
        self.score = score
        self.total_players = total_players

    @classmethod
    def _from_dict(cls, raw: Dict):
        _log.debug("Creating a PlayerCOTDResults class from given dictionary")

        id = raw.get("id")
        timestamp = datetime.strptime(raw.get("timestamp"), "%Y-%m-%dT%H:%M:%S+00:00")
        name = raw.get("name")
        div = raw.get("div")
        rank = raw.get("rank")
        if raw.get("divrank") != 0:
            div_rank = raw.get("divrank")
            score = raw.get("score")
        else:
            div_rank = score = 0
        total_players = raw.get("totalplayers")

        return cls(
            id,
            timestamp,
            name,
            div,
            rank,
            div_rank,
            score,
            total_players,
        )


class PlayerCOTD:
    """
    .. versionadded :: 0.3.0

    The Player's COTD Data

    Parameters
    ----------
    total : int
        Total COTD's Played
    recent_results : :class:`List[PlayerCOTDResults]`
        Represents the recent COTD results the player has gotten
    stats : :class:`PlayerCOTDStats`
        Represents the Statistics of the Player's COTD career
    player_id : str
        The player's ID
    """

    def __init__(
        self,
        total: int,
        recent_results: List[PlayerCOTDResults],
        stats: PlayerCOTDStats,
        player_id: str,
    ):
        self.total = total
        self.recent_results = recent_results
        self.stats = stats
        self.player_id = player_id

    @classmethod
    def _from_dict(cls, page_data: Dict, player_id: str):
        _log.debug("Creating a PlayerCOTD class from given dictionary")

        total = page_data.get("total")
        stats = PlayerCOTDStats._from_dict(page_data.get("stats"))
        player_id = player_id

        recent_results = []
        for cotd in page_data.get("cotds"):
            recent_results.append(PlayerCOTDResults._from_dict(cotd))

        return cls(
            total,
            recent_results,
            stats,
            player_id,
        )

    @classmethod
    async def get_page(cls: Self, player_id: str, page: int = 0) -> Self:
        """
        .. versionadded :: 0.3.0

        Gets the Player's COTD Stats of a particular page.

        Parameters
        ----------
        player_id : str
            The player's ID
        page : int, optional
            The page of the ID, by default 0
        """
        return cls._from_dict(await _get_trophy_page(player_id, page), player_id)


class COTD:
    """
    .. versionadded :: 0.3.0

    Represents a Cup of the Day

    Parameters
    ----------
    cotd_id : int
        The cotd id
    name : str
        The name of the cotd
    player_count : int
        The number of players that played the :class:`COTD`
    start_date : :class:`datetime`
        The start date of the COTD
    end_date : :class:`datetime`
        The end date of the COTD
    """

    def __init__(
        self,
        cotd_id: int,
        name: str,
        player_count: int,
        start_date: datetime,
        end_date: datetime,
    ):
        self.cotd_id = cotd_id
        self.name = name
        self.player_count = player_count
        self.start_date = start_date
        self.end_date = end_date

    @classmethod
    def _from_dict(cls, raw: Dict):
        _log.debug("Creating a COTD class from given dictionary")

        cotd_id = raw.get("id")
        name = raw.get("name")
        player_count = raw.get("players")
        start_date = datetime.utcfromtimestamp(raw.get("starttime"))
        end_date = datetime.utcfromtimestamp(raw.get("endtime"))

        return cls(
            cotd_id,
            name,
            player_count,
            start_date,
            end_date,
        )

    @classmethod
    async def get_cotd(cls: Self, page: int = 0) -> List[Self]:
        """
        .. versionadded :: 0.3.0

        Fetches the Latest COTDs and returns its data.

        Parameters
        ----------
        page : int, optional
            The page, each page contains 12 items. by default 0

        Returns
        -------
        :class:`List[COTD]`
            The COTDs
        """
        all_cotds = await _get_cotd_page(page)

        acotd_fmt = []
        for cotd in all_cotds:
            acotd_fmt.append(cls._from_dict(cotd))

        return acotd_fmt



import json
import logging
from contextlib import suppress
from datetime import datetime
from typing import Dict

import redis
from typing_extensions import Self

from trackmania.errors import InvalidTOTDDate, TMIOException

from .api import _APIClient
from .config import Client
from .constants import _TMIO
from .errors import TMIOException, TrackmaniaException
from .tmmap import TMMap

_log = logging.getLogger(__name__)

__all__ = ("TOTD",)


class TOTD:
    """
    .. versionadded :: 0.3.0

    Class that represents a TOTD

    Parameters
    ----------
    campaign_id : int
        The campaign's id
    leaderboard_uid : int
        The leaderboard's uid
    month_day : int
        The day of the month when the totd was played
    week_day : int
        The day of the week when the totd was played
    map : :class:`TMMap`
        The map that was played
    """

    def __init__(
        self,
        campaign_id: int,
        leaderboard_uid: int,
        month_day: int,
        week_day: int,
        mapobj: TMMap,
    ):
        self.campaign_id = campaign_id
        self.leaderboard_uid = leaderboard_uid
        self.month_day = month_day
        self.week_day = week_day
        self._mapobj = mapobj

    @classmethod
    def _from_dict(cls, raw: Dict):
        campaign_id = raw.get("campaignid")
        mapobj = TMMap._from_dict(raw.get("map"))
        week_day = raw.get("weekday")
        month_day = raw.get("monthday")
        leaderboard_uid = raw.get("leaderboarduid")

        return cls(
            campaign_id,
            leaderboard_uid,
            month_day,
            week_day,
            mapobj,
        )

    @staticmethod
    def _calculate_months(date: datetime) -> int:
        """
        .. versionadded :: 0.3.0

        Calculates the number of months from the given date to the current month.

        Parameters
        ----------
        date : datetime
            The date to calculate to

        Returns
        -------
        int
            How many months it has been
        """
        today = datetime.utcnow()
        today_month = today.month
        today_year = today.year

        months = (date.year - today_year) * 12
        return (months + date.month - today_month) * -1

    @property
    def map(self):
        """TMMap Property"""
        return self._mapobj

    @classmethod
    async def get_totd(cls: Self, date: datetime) -> Self:
        """
        .. versionadded :: 0.3.0

        Gets a map from the date provided.

        Parameters
        ----------
        date : datetime
            The date of the TOTD.

        Returns
        -------
        :class:`TOTD`
            The map
        """
        _log.debug("Getting TOTD for date: %s", date)

        cache_client = Client._get_cache_client()
        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            if cache_client.exists(f"totd:{date.year}:{date.month}:{date.day}"):
                _log.debug(
                    f"Found TOTD for date {date.day}:{date.month}:{date.year} in cache"
                )
                return cls._from_dict(
                    json.loads(
                        cache_client.get(
                            f"totd:{date.year}:{date.month}:{date.day}"
                        ).decode("utf-8")
                    )
                )

        api_client = _APIClient()
        all_totds = await api_client.get(
            _TMIO.build([_TMIO.TABS.TOTD, TOTD._calculate_months(date)])
        )
        await api_client.close()

        with suppress(KeyError, TypeError):
            raise TMIOException(all_totds["error"])

        if all_totds["lastday"] < date.day:
            raise InvalidTOTDDate(
                f"The date provided is not a valid TOTD date. The last day is {all_totds['lastday']}"
            )

        try:
            totd = all_totds["days"][date.day - 1]
        except (IndexError) as excp:
            raise InvalidTOTDDate("That TOTD Date is not correct.") from excp
        except (KeyError, TypeError) as excp:
            raise TrackmaniaException(
                f"Something Unexpected has occured. Please contact the developer of the Package.\nMessage: {excp}"
            ) from excp

        with suppress(ConnectionRefusedError, redis.exceptions.ConnectionError):
            _log.debug(f"Caching TOTD for date {date.day}:{date.month}:{date.year}")
            cache_client.set(
                f"totd:{date.year}:{date.month}:{date.day}", json.dumps(totd)
            )

        return cls._from_dict(totd)

    @classmethod
    async def latest_totd(cls: Self) -> Self:
        """
        .. versionadded :: 0.3.3

        Gets the latest totd.

        Returns
        -------
        :class:`TOTD`
            The TOTD object.
        """
        today = datetime.utcnow()

        if today.hour > 17 and today.minute > 0:
            return await cls.get_totd(datetime.utcnow())
        else:
            return await cls.get_totd(datetime(today.year, today.month, today.day - 1))



from typing import List

from typing_extensions import Self


class _TMIOTabs:
    """
    .. versionadded:: 0.3.0

    _TMIO Endpoints
    """

    def __init__(self):
        self.PLAYER: str = "player"
        self.PLAYERS: str = "players"
        self.TROPHIES: str = "trophies"
        self.MAP: str = "map"
        self.LEADERBOARD: str = "leaderboard"
        self.MATCHES: str = "matches"

        self.TOP_MATCHMAKING: str = "top/matchmaking/2"
        self.TOP_ROYAL: str = "top/matchmaking/3"
        self.TOP_TROPHIES: str = "top/trophies"

        self.MATCHMAKING_ID: int = 2
        self.ROYAL_ID: int = 3

        self.TOTD: str = "totd"
        self.COTD: str = "cotd"
        self.ADS: str = "ads"


class _TMIO:
    """
    .. versionadded:: 0.3.0

    Basic _TMIO Api Details

    Parameters
    ----------
    PROTOCOL : str
        The PROTOCOL to use for the api. Equal to "https"
    BASE : str
        The BASE url for the api. Equal to "trackmania.io"
    api : str
        The api endpoint for `trackmania.io`. Equal to "api".
    TABS : :class:`_TMIOTabs`
        The TABS for the api.
    """

    PROTOCOL: str = "https"
    BASE: str = "trackmania.io"
    API: str = "api"

    TABS: _TMIOTabs = _TMIOTabs()

    @classmethod
    def build(cls: Self, endpoints: List[str]) -> str:
        """Builds a _TMIO endpoint url.

        Parameters
        ----------
        endpoints : List[str]
            The endpoints to build the url with.

        Returns
        -------
        str
            The built endpoint url.

        """
        url = f"{cls.PROTOCOL}://{cls.BASE}/{cls.API}/"

        if len(endpoints) == 0:
            return url
        if len(endpoints) > 1:
            for item in endpoints:
                if isinstance(item, int):
                    item = str(item)
                url = url + item + "/"

            # Removing the Final /
            return url[:-1]
        return url + endpoints[0]


class _TMXTabs:
    """
    .. versionadded :: 0.3.3

    _TMX Endpoints
    """

    def __init__(self):
        self.MAPS = "maps"
        self.GET_MAP_INFO = "get_map_info"
        self.GET_TRACK_INFO = "get_track_info"
        self.MULTI = "multi"
        self.ID = "id"


class _TMX:
    """
    .. versionadded:: 0.3.0

    Basic _TMX Api Details

    Parameters
    ----------
    PROTOCOL : str
        The PROTOCOL to use for the api. Equal to "https"
    BASE : str
        The BASE url for the api. Equal to "trackmania.exchange"
    api : str
        The api endpoint for `trackmania.exchange`. Equal to "api".
    TABS: :class:`_TMXTabs`
        .. versionadded :: 0.3.3
        The TABS for TMX API
    """

    PROTOCOL: str = "https"
    BASE: str = "trackmania.exchange"
    API: str = "api"
    TABS: _TMXTabs = _TMXTabs()

    MAP_TYPE_ENUMS: dict = {
        1: "Race",
        2: "FullSpeed",
        3: "Tech",
        4: "RPG",
        5: "LOL",
        6: "Press Forward",
        7: "SpeedTech",
        8: "MultiLap",
        9: "Offroad",
        10: "Trial",
        11: "ZrT",
        12: "SpeedFun",
        13: "Competitive",
        14: "Ice",
        15: "Dirt",
        16: "Stunt",
        17: "Reactor",
        18: "Platform",
        19: "Slow Motion",
        20: "Bumper",
        21: "Fragile",
        22: "Scenery",
        23: "Kacky",
        24: "Endurance",
        25: "Mini",
        26: "Remake",
        27: "Mixed",
        28: "Nascar",
        29: "SpeedDrift",
        30: "Minigame",
        31: "Obstacle",
        32: "Transitional",
        33: "Grass",
        34: "Backwards",
        35: "Freewheel",
        36: "Signature",
        37: "Royal",
        38: "Water",
        39: "Plastic",
        40: "Arena",
    }

    @classmethod
    def build(cls: Self, endpoints: List[str]) -> str:
        """URL Builder for _TMX API

        Parameters
        ----------
        endpoints : class:`List`[str]
            The endpoints as a list.

        Returns
        -------
        str
            The URL.
        """
        url = f"{cls.PROTOCOL}://{cls.BASE}/{cls.API}/"

        if len(endpoints) == 0:
            return url
        if len(endpoints) > 1:
            for item in endpoints:
                if isinstance(item, int):
                    item = str(item)
                url = url + item + "/"

            # Removing the Final /
            return url[:-1]
        return url + endpoints[0]



from setuptools import setup

requirements = ["aiohttp", "redis"]

version = "v0.4.0b1"

readme = ""
with open("README.md") as f:
    readme = f.read()

packages = [
    "trackmania",
]

setup(
    name="py-tmio",
    author="Deepesh Nimma",
    url="https://github.com/NottCurious/py-tmio",
    project_urls={
        "Documentation": "https://py-trackmaniaio.rtfd.io/",
        "Issue Tracker": "https://github.com/NottCurious/py-tmio/issues",
    },
    version=version,
    license="MIT",
    description="Trackmania.io API Wrapper",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=requirements,
    python_requires=">=3.10",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=packages,
)



import redis

__all__ = ("Client",)


class Client:
    """
    .. versionadded:: 0.3.0

    Client class to manage user defined constants

    Parameters
    ----------
    USER_AGENT : str
        The USER_AGENT to be used for the bot.
    REDIS_HOST : str
        The host of the redis server.
    REDIS_PORT : int
        The port of the redis server.
    REDIS_DB : int
        The database of the redis server.
        .. versionadded:: 0.2.0
    REDIS_PASSWORD : str
        The password of the redis server.
        .. versionadded:: 0.2.0
    RATELIMIT_LIMIT: int
        The `trackmania.io` ratelimit limit.
        .. versionadded:: 0.2.1
    RATELIMIT_REMAINING : int
        The amount of remaining requests with `trackmania.io` api.
        .. versionadded:: 0.2.1
    """

    USER_AGENT: str = None
    REDIS_HOST: str = "127.0.0.1"
    REDIS_PORT: int = 6379
    REDIS_DB: int = 0
    REDIS_PASSWORD: str = None

    RATELIMIT_LIMIT: int = 40
    RATELIMIT_REMAINING: int = None

    @staticmethod
    def _get_cache_client() -> redis.Redis:
        """
        Gets the Cache Client

        Returns
        -------
        :class:`redis.Redis`
            The cache_client
        """
        return redis.Redis(
            host=Client.REDIS_HOST,
            port=Client.REDIS_PORT,
            db=Client.REDIS_DB,
            password=Client.REDIS_PASSWORD,
        )



import logging
from typing import Optional

import aiohttp

from .config import Client
from .errors import NoUserAgentSetError

__all__ = ("ResponseCodeError", "_APIClient")
_log = logging.getLogger(__name__)


class ResponseCodeError(ValueError):
    """
    .. versionadded:: 0.3.0

    Raised when a non-OK HTTP Response is received

    Parameters
    ----------
    response : :class:`aiohttp.ClientResponse`
        The response object
    response_json : :class:`Dict`
        The response json
    response_text : str
        The response text

    """

    def __init__(
        self,
        response: aiohttp.ClientResponse,
        response_json: Optional[dict] = None,
        response_text: str = "",
    ):
        super().__init__(response_text)
        self.status = response.status
        self.response_json = response_json or {}
        self.response_text = response_text
        self.response = response

    def __str__(self):
        response = self.response_json if self.response_json else self.response_text
        return f"Status: {self.status} Response: {response}"


# pylint: disable=W0612
class _APIClient:
    """
    .. versionadded:: 0.3.0

    API Wrappers
    """

    def __init__(self, **session_kwargs):
        if Client.USER_AGENT is None:
            raise NoUserAgentSetError()

        self.session = aiohttp.ClientSession(
            headers={"User-Agent": Client.USER_AGENT + " | via py-tmio"},
            **session_kwargs,
        )

    async def close(self) -> None:
        """
        Close the AIOHTTP Session
        """

        await self.session.close()

    # pylint: disable=R0201
    async def maybe_raise_for_status(
        self, response: aiohttp.ClientResponse, should_raise: bool
    ) -> None:
        """Raise ResponseCodeError for non-OK response if an exception should be raised"""
        if should_raise and response.status >= 400:
            try:
                response_json = await response.json()
                if "error" in response_json:
                    return
                await self.session.close()
                raise ResponseCodeError(response=response, response_json=response_json)
            except aiohttp.ContentTypeError as content_type_error:
                response_text = await response.text()
                if "error" in response_text:
                    return
                await self.session.close()
                raise ResponseCodeError(
                    response=response, response_text=response_text
                ) from content_type_error

    async def request(
        self,
        method: str,
        endpoint: str,
        *,
        raise_for_status: bool = True,
        **kwargs,
    ) -> dict:
        """Send an HTTP request to the site API and return the JSON response."""
        async with self.session.request(method.upper(), endpoint, **kwargs) as resp:
            await self.maybe_raise_for_status(resp, raise_for_status)
            _log.info(f"Sending {method.upper()} to {endpoint}")
            try:
                if "trackmania.io" in endpoint:
                    Client.RATELIMIT_LIMIT = int(
                        resp.headers.get("x-ratelimit-limit")[0]
                    )
                    Client.RATELIMIT_REMAINING = int(
                        resp.headers.get("x-ratelimit-remaining")[0]
                    )
            except (AttributeError, TypeError):
                pass
            try:
                return await resp.json()
            except:
                return await resp.text()

    async def get(
        self,
        endpoint: str,
        *,
        raise_for_status: bool = True,
        **kwargs,
    ) -> dict:
        """Site API GET."""
        return await self.request(
            "GET",
            endpoint,
            raise_for_status=raise_for_status,
            **kwargs,
        )

    async def patch(
        self, endpoint: str, *, raise_for_status: bool = True, **kwargs
    ) -> dict:
        """Site API PATCH."""
        return await self.request(
            "PATCH", endpoint, raise_for_status=raise_for_status, **kwargs
        )

    async def post(
        self, endpoint: str, *, raise_for_status: bool = True, **kwargs
    ) -> dict:
        """Site API POST."""
        return await self.request(
            "POST", endpoint, raise_for_status=raise_for_status, **kwargs
        )

    async def put(
        self, endpoint: str, *, raise_for_status: bool = True, **kwargs
    ) -> dict:
        """Site API PUT."""
        return await self.request(
            "PUT", endpoint, raise_for_status=raise_for_status, **kwargs
        )
